/*
 * ShadowIRCd - an advanced Internet Relay Chat daemon(ircd)
 * simban.c: A simban implementation.
 * 
 * Copyright (c) 2004 ShadowIRCd development group.
 * 
 * For licensing details, see LICENSE in the root of this source.
 *
 * $Id: simban.c,v 1.1.2.27 2004/11/08 01:38:51 nenolod Exp $
 */

#include "stdinc.h"
#include "common.h"
#include "tools.h"
#include "s_log.h"
#include "s_conf.h"
#include "s_misc.h"
#include "hostmask.h"
#include "client.h"
#include "channel.h"
#include "reject.h"
#include "irc_string.h"
#include "sprintf_irc.h"
#include "memory.h"
#include "send.h"
#include "resv.h"
#include "hook.h"
#include "s_serv.h"
#include "list.h"
#include "simban.h"
#include "balloc.h"
#include "ircd.h"

dlink_list simban_list;
void parse_simban(char *);
static void clear_simbans();
static int tokenize_line(char *, char **);
static int simban_lineno;
void onjoin_simban_check(struct Membership *);

char result[BUFSIZE];

struct SimbanConfigToken {
	char *name;
	int value;
};

struct SimbanConfigToken SimbanTargets[] = {
	{"reject", SIMBAN_REJECT},
	{"except", SIMBAN_EXEMPT},
	{"autoban", SIMBAN_AUTOBAN},
	{"allow", SIMBAN_ALLOW},
	{NULL, 0}
};

struct SimbanConfigToken SimbanRules[] = {
	{"nick", SIMBAN_NICK},
	{"gecos", SIMBAN_GECOS},
	{"channel", SIMBAN_CHANNEL},
	{"hosts", SIMBAN_HOSTS},
	{"ip", SIMBAN_IP},
	{"privmsg", SIMBAN_PRIVMSG},
	{NULL, 0}
};

struct SimbanConfigToken SimbanFlags[] = {
	{"matching", SIMBAN_MATCHING},
	{"regex", SIMBAN_REGEX},
	{NULL, 0}
};

/* initializes the simban heap table */
void
init_simbans()
{
	/*
	 * put in future initialization code here 
	 */
	return;
}

/*
 * parse_simbans - loads the simbans file in and parses it
 * 
 * input	- none
 * output	- none
 * side effects - various simbans are loaded into the ircd
 */
void
parse_simbans()
{
	FBFILE *file;
	char line[BUFSIZE];
	char *p;

	clear_simbans();

	simban_lineno = 1;

	if ((file = fbopen(SBPATH, "r")) == NULL) {
		ilog(L_ERROR, "Failed reading file %s", SBPATH);
		sendto_realops_flags(UMODE_ALL, L_ALL, "Can't open %s file",
		    SBPATH);
	} else {
		while (fbgets(line, sizeof(line), file) != NULL) {
			if ((p = strchr(line, '\n')) != NULL)
				*p = '\0';

			if ((*line == '\0') || (*line == '#')) {
				simban_lineno++;
				continue;
			}
			parse_simban(line);
			simban_lineno++;
		}
		fbclose(file);
	}
}

/* tokenizes a line. used internally. don't touch this. */
static int
tokenize_line(char *line, char *parv[9])
{
	char *p;
	char *buf = line;
	int x = 0;

	parv[x] = NULL;

	while (*buf == ' ')	/* skip leading spaces */
		buf++;

	if (*buf == '\0')	/* ignore all-space args */
		return (x);

	do {
		if (*buf == ':') {	/* Last parameter */
			buf++;
			parv[x++] = buf;
			parv[x] = NULL;
			return (x);
		} else {
			parv[x++] = buf;
			parv[x] = NULL;

			if ((p = strchr(buf, ' ')) != NULL) {
				*p++ = '\0';
				buf = p;
			} else
				return (x);
		}

		while (*buf == ' ')
			buf++;

		if (*buf == '\0')
			return (x);
	}
	while (x < 8);

	return (x);
}

static void
clear_simbans()
{
	dlink_node *node, *next;
	struct simban *simban;

	DLINK_FOREACH_SAFE(node, next, simban_list.head) {
		simban = node->data;

		dlinkDelete(node, &simban_list);

		memset(simban, 0, sizeof(struct simban));
		MyFree(simban);
		free_dlink_node(node);
	}
}

char *
print_simban(struct simban *simban)
{
	char *rule, *class, *flags;
	int i;

	for (i = 0; SimbanTargets[i].name; i++) {
		if (simban->ruletype == SimbanTargets[i].value)
			DupString(rule, SimbanTargets[i].name);
	}

	for (i = 0; SimbanRules[i].name; i++) {
		if (simban->ruleclass == SimbanRules[i].value)
			DupString(class, SimbanRules[i].name);
	}

	for (i = 0; SimbanFlags[i].name; i++) {
		if (simban->flags == SimbanRules[i].value)
			DupString(flags, SimbanRules[i].name);
	}

	if (simban->flags) {
		ircsprintf(result, "%s %s %s %s", rule, class, flags,
		    simban->param);
	} else {
		ircsprintf(result, "%s %s %s", rule, class, simban->param);
	}

	return result;
}

void
report_simbans(struct Client *source_p)
{
	dlink_node *node;
	struct simban *simban;

	DLINK_FOREACH(node, simban_list.head) {
		simban = node->data;

		sendto_one(source_p, ":%s 218 %s %s",
		    me.name, source_p->name, print_simban(simban));
	}
}

void
parse_simban(char *line)
{
	char *tokens[9];
	dlink_node *node = make_dlink_node();
	struct simban *simban = MyMalloc(sizeof(struct simban));
	int paramslot = 0;
	int i;
	int ii;

	tokenize_line(line, tokens);

	/*
	 * the first token will always be the type of rule... that means, it
	 * will always be reject, except, autoban or allow
	 */
	for (i = 0; SimbanTargets[i].name; i++) {
		if (!strcasecmp(SimbanTargets[i].name, tokens[0]))
			simban->ruletype = SimbanTargets[i].value;
	}

	/*
	 * the second token will always be the class. 
	 */
	for (i = 0; SimbanRules[i].name; i++) {
		if (!strcasecmp(SimbanRules[i].name, tokens[1]))
			simban->ruleclass = SimbanRules[i].value;
	}

	/*
	 * this finds the param token, assuming it is the last token 
	 */
	for (paramslot = 0; tokens[paramslot]; paramslot++);

	paramslot--;

	/*
	 * now we scan tokens between the two points 
	 */
	if (paramslot > 2) {
		for (i = 2; i <= paramslot; i++) {
			for (ii = 0; SimbanFlags[ii].name; ii++) {
				if (!strcasecmp(SimbanRules[ii].name,
					tokens[i]))
					simban->flags = SimbanRules[ii].value;
			}
		}
	}

	strcpy(simban->param, tokens[paramslot]);

	dlinkAdd(simban, node, &simban_list);
}

static void
do_autoban(struct Client *client_p, struct simban *simban)
{
	struct AccessItem *aconf;
	struct ConfItem *conf;

	conf = make_conf_item(KLINE_TYPE);
	aconf = (struct AccessItem *)map_to_conf(conf);
	DupString(aconf->host, client_p->host);
	DupString(aconf->user, "*");
	aconf->port = 0;
	DupString(aconf->reason, "Autobanned");

	add_conf_by_address(CONF_KILL, aconf);

	/*
	 * If we have general::permanant_simbans set to yes, add this ban to
	 * the kline file
	 */
	if (ConfigFileEntry.permanant_simbans)
		write_conf_line(client_p, conf, smalldate(CurrentTime),
		    CurrentTime);

	/*
	 * Enable kline checking down the line 
	 */
	rehashed_klines = 1;
}

int
onconnect_simban_check(struct Client *client_p)
{
	dlink_node *node;
	struct simban *simban;

	DLINK_FOREACH(node, simban_list.head) {
		simban = node->data;

		if (!strcasecmp(client_p->host, simban->param)
		    && simban->ruleclass == SIMBAN_HOSTS) {
			/*
			 * We have a match. What do we do here? 
			 */
			if (simban->ruletype == SIMBAN_REJECT) {
				add_reject(client_p);
				exit_client(client_p, client_p, client_p,
				    "Rejected");
				sendto_realops_flags(UMODE_CCONN, L_ALL,
				    "Rejecting client: %s (%s@%s) via simban rule: %s",
				    client_p->name, client_p->username,
				    client_p->host, print_simban(simban));
				return 1;
			}
			if (simban->ruletype == SIMBAN_AUTOBAN) {
				do_autoban(client_p, simban);
				return 0;
			}
			if (simban->ruletype == SIMBAN_EXEMPT) {
				sendto_one(client_p,
				    ":%s NOTICE %s :*** You're immune to simban checks, aren't you fearsome?",
				    me.name, client_p->name);
				return 0;
			}
		} else if (match(simban->param, client_p->host) &&
		    simban->ruleclass == SIMBAN_HOSTS
		    && simban->flags == SIMBAN_MATCHING) {
			/*
			 * We have a fuzzy match. 
			 */
			if (simban->ruletype == SIMBAN_REJECT) {
				add_reject(client_p);
				exit_client(client_p, client_p, client_p,
				    "Rejected");
				sendto_realops_flags(UMODE_CCONN, L_ALL,
				    "Rejecting client: %s (%s@%s) via simban rule: %s",
				    client_p->name, client_p->username,
				    client_p->host, print_simban(simban));
				return 1;
			}
			if (simban->ruletype == SIMBAN_AUTOBAN) {
				do_autoban(client_p, simban);
				return 0;
			}
			if (simban->ruletype == SIMBAN_EXEMPT) {
				sendto_one(client_p,
				    ":%s NOTICE %s :*** You're immune to simban checks, aren't you fearsome?",
				    me.name, client_p->name);
				return 0;
			}
		}
		if (!strcasecmp(client_p->name, simban->param)
		    && simban->ruleclass == SIMBAN_NICK) {
			/*
			 * We have a match. What do we do here? 
			 */
			if (simban->ruletype == SIMBAN_REJECT) {
				add_reject(client_p);
				exit_client(client_p, client_p, client_p,
				    "Rejected");
				sendto_realops_flags(UMODE_CCONN, L_ALL,
				    "Rejecting client: %s (%s@%s) via simban rule: %s",
				    client_p->name, client_p->username,
				    client_p->host, print_simban(simban));
				return 1;
			}
			if (simban->ruletype == SIMBAN_AUTOBAN) {
				do_autoban(client_p, simban);
				return 0;
			}
			if (simban->ruletype == SIMBAN_EXEMPT) {
				sendto_one(client_p,
				    ":%s NOTICE %s :*** You're immune to simban checks, aren't you fearsome?",
				    me.name, client_p->name);
				return 0;
			}
		} else if (match(simban->param, client_p->name) &&
		    simban->ruleclass == SIMBAN_NICK
		    && simban->flags == SIMBAN_MATCHING) {
			/*
			 * We have a fuzzy match. 
			 */
			if (simban->ruletype == SIMBAN_REJECT) {
				add_reject(client_p);
				exit_client(client_p, client_p, client_p,
				    "Rejected");
				return 1;
			}
			if (simban->ruletype == SIMBAN_AUTOBAN) {
				do_autoban(client_p, simban);
				return 0;
			}
			if (simban->ruletype == SIMBAN_EXEMPT) {
				sendto_one(client_p,
				    ":%s NOTICE %s :*** You're immune to simban checks, aren't you fearsome?",
				    me.name, client_p->name);
				return 0;
			}
		}
		if (!strcasecmp(client_p->info, simban->param)
		    && simban->ruleclass == SIMBAN_GECOS) {
			/*
			 * We have a match. What do we do here? 
			 */
			if (simban->ruletype == SIMBAN_REJECT) {
				add_reject(client_p);
				exit_client(client_p, client_p, client_p,
				    "Rejected");
				sendto_realops_flags(UMODE_CCONN, L_ALL,
				    "Rejecting client: %s (%s@%s) via simban rule: %s",
				    client_p->name, client_p->username,
				    client_p->host, print_simban(simban));
				return 1;
			}
			if (simban->ruletype == SIMBAN_AUTOBAN) {
				do_autoban(client_p, simban);
				return 0;
			}
			if (simban->ruletype == SIMBAN_EXEMPT) {
				sendto_one(client_p,
				    ":%s NOTICE %s :*** You're immune to simban checks, aren't you fearsome?",
				    me.name, client_p->name);
				return 0;
			}
		} else if (match(simban->param, client_p->info) &&
		    simban->ruleclass == SIMBAN_GECOS
		    && simban->flags == SIMBAN_MATCHING) {
			/*
			 * We have a fuzzy match. 
			 */
			if (simban->ruletype == SIMBAN_REJECT) {
				add_reject(client_p);
				exit_client(client_p, client_p, client_p,
				    "Rejected");
				sendto_realops_flags(UMODE_CCONN, L_ALL,
				    "Rejecting client: %s (%s@%s) via simban rule: %s",
				    client_p->name, client_p->username,
				    client_p->host, print_simban(simban));
				return 1;
			}
			if (simban->ruletype == SIMBAN_AUTOBAN) {
				do_autoban(client_p, simban);
				return 0;
			}
			if (simban->ruletype == SIMBAN_EXEMPT) {
				sendto_one(client_p,
				    ":%s NOTICE %s :*** You're immune to simban checks, aren't you fearsome?",
				    me.name, client_p->name);
				return 0;
			}
		} else if (simban->flags == SIMBAN_REGEX) {
			pcre *regex;
			int *offset;

			regex =
			    pcre_compile(simban->param, PCRE_CASELESS, NULL,
			    offset, NULL);
			if (pcre_exec(regex, NULL, client_p->info,
				strlen(client_p->info), 0, PCRE_CASELESS, NULL,
				0)) {
				if (simban->ruletype == SIMBAN_REJECT)
					return 1;
				if (simban->ruletype == SIMBAN_AUTOBAN) {
					do_autoban(client_p, simban);
					return 1;
				}
			}
		}
		if (!strcasecmp(client_p->ipaddr, simban->param)
		    && simban->ruleclass == SIMBAN_IP) {
			/*
			 * We have a match. What do we do here? 
			 */
			if (simban->ruletype == SIMBAN_REJECT) {
				add_reject(client_p);
				exit_client(client_p, client_p, client_p,
				    "Rejected");
				sendto_realops_flags(UMODE_CCONN, L_ALL,
				    "Rejecting client: %s (%s@%s) via simban rule: %s",
				    client_p->name, client_p->username,
				    client_p->host, print_simban(simban));
				return 1;
			}
			if (simban->ruletype == SIMBAN_AUTOBAN) {
				do_autoban(client_p, simban);
				return 0;
			}
			if (simban->ruletype == SIMBAN_EXEMPT) {
				sendto_one(client_p,
				    ":%s NOTICE %s :*** You're immune to simban checks, aren't you fearsome?",
				    me.name, client_p->name);
				return 0;
			}
		} else if (match(simban->param, client_p->ipaddr) &&
		    simban->ruleclass == SIMBAN_IP
		    && simban->flags == SIMBAN_MATCHING) {
			/*
			 * We have a fuzzy match. 
			 */
			if (simban->ruletype == SIMBAN_REJECT) {
				add_reject(client_p);
				exit_client(client_p, client_p, client_p,
				    "Rejected");
				sendto_realops_flags(UMODE_CCONN, L_ALL,
				    "Rejecting client: %s (%s@%s) via simban rule: %s",
				    client_p->name, client_p->username,
				    client_p->host, print_simban(simban));
				return 1;
			}
			if (simban->ruletype == SIMBAN_AUTOBAN) {
				do_autoban(client_p, simban);
				return 0;
			}
			if (simban->ruletype == SIMBAN_EXEMPT) {
				sendto_one(client_p,
				    ":%s NOTICE %s :*** You're immune to simban checks, aren't you fearsome?",
				    me.name, client_p->name);
				return 0;
			}
		}
	}

	return 0;
}

int
onprivmsg_simban_check(struct Client *client_p, struct Client *target_p,
    const char *msg)
{
	dlink_node *node;
	struct simban *simban;

	if (!MyClient(client_p))
		return 0;

	DLINK_FOREACH(node, simban_list.head) {
		simban = node->data;

		if (!strcasecmp(msg, simban->param) &&
		    simban->ruleclass == SIMBAN_PRIVMSG) {
			/*
			 * We have a match. 
			 */
			if (simban->ruletype == SIMBAN_REJECT) {
				/*
				 * stop the message from being sent 
				 */
				return 1;
			}
			if (simban->ruletype == SIMBAN_AUTOBAN) {
				do_autoban(client_p, simban);
				return 1;
			}
		} else if (match(simban->param, msg) &&
		    simban->ruleclass == SIMBAN_PRIVMSG &&
		    simban->flags == SIMBAN_MATCHING) {
			printf("match!\n");
			if (simban->ruletype == SIMBAN_REJECT) {
				return 1;
			}
			if (simban->ruletype == SIMBAN_AUTOBAN) {
				do_autoban(client_p, simban);
				return 1;
			}
		} else if (simban->flags == SIMBAN_REGEX) {
			pcre *regex;
			int *offset;

			regex =
			    pcre_compile(simban->param, PCRE_CASELESS, NULL,
			    offset, NULL);
			if (pcre_exec(regex, NULL, msg, strlen(msg), 0,
				PCRE_CASELESS, NULL, 0)) {
				if (simban->ruletype == SIMBAN_REJECT)
					return 1;
				if (simban->ruletype == SIMBAN_AUTOBAN) {
					do_autoban(client_p, simban);
					return 1;
				}
			}
		}
	}

	return 0;
}

void
onjoin_simban_check(struct Membership *ms)
{
	dlink_node *node;
	struct simban *simban;
	struct Client *client_p;

	client_p = ms->client_p;

	if (!MyClient(client_p))
		return;

	DLINK_FOREACH(node, simban_list.head) {
		simban = node->data;

		if (!strcasecmp(ms->chptr->chname, simban->param) &&
		    simban->ruleclass == SIMBAN_CHANNEL) {
			/*
			 * We have a match. What do we do here? 
			 */
			if (simban->ruletype == SIMBAN_REJECT) {
				add_reject(client_p);
				exit_client(client_p, client_p, client_p,
				    "Rejected");
				sendto_realops_flags(UMODE_CCONN, L_ALL,
				    "Rejecting client: %s (%s@%s) via simban rule: %s",
				    client_p->name, client_p->username,
				    client_p->host, print_simban(simban));
				return;
			}
			if (simban->ruletype == SIMBAN_AUTOBAN) {
				do_autoban(client_p, simban);
				return;
			}
			if (simban->ruletype == SIMBAN_EXEMPT) {
				return;
			}
		} else if (match(simban->param, ms->chptr->chname) &&
		    simban->ruleclass == SIMBAN_CHANNEL
		    && simban->flags == SIMBAN_MATCHING) {
			/*
			 * We have a fuzzy match. 
			 */
			if (simban->ruletype == SIMBAN_REJECT) {
				add_reject(client_p);
				exit_client(client_p, client_p, client_p,
				    "Rejected");
				sendto_realops_flags(UMODE_CCONN, L_ALL,
				    "Rejecting client: %s (%s@%s) via simban rule: %s",
				    client_p->name, client_p->username,
				    client_p->host, print_simban(simban));
				return;
			}
			if (simban->ruletype == SIMBAN_AUTOBAN) {
				do_autoban(client_p, simban);
				return;
			}
			if (simban->ruletype == SIMBAN_EXEMPT) {
				return;
			}
		}
	}
}
